// Ovo je samo jedno od mnogih mogućih rešenja zadatka

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#define NUM_THREADS 5

void create_and_load_array(int **numbers, int *count_addr)
{
	int i;
	printf("Unesite broj elemenata niza: ");
	scanf("%d", count_addr);
	*numbers = calloc(*count_addr, sizeof(int));
	if (NULL != *numbers && *count_addr >= NUM_THREADS)
	{
		printf("Unesite elemente niza: ");
		for (i = 0; i < *count_addr; i++)
		{
			scanf("%d", *numbers + i);
		}
	}
	else
	{
		*count_addr = 0;
	}
}

typedef struct
{
	int minimum;
	int maximum;
} Results;

Results results;

typedef struct 
{
	int *elements;
	unsigned base;
	unsigned end_offset;
	pthread_mutex_t *lock_addr;
} ThreadData;

void print_results(Results results)
{
	printf("Min: %d, max: %d\n", results.minimum, results.maximum);
}

void* work_routine(void *data)
{
//  sve što je uslovljeno sa PTW32_STATIC_LIB je potrebno 
//  samo za pthreads-win32 i to samo u static linked varijanti
#ifdef PTW32_STATIC_LIB
	pthread_win32_thread_attach_np();
	{
#endif
	unsigned i; 
	int localMin, localMax;
	ThreadData *thread_data = (ThreadData*) data;

	// kvadriraj elemente i traži lokalni maksimum
	localMin = localMax = thread_data->elements[thread_data->base]*thread_data->elements[thread_data->base];
	for (i = thread_data->base; i < thread_data->end_offset; i++)
	{
		thread_data->elements[i] *= thread_data->elements[i];
		if (localMin > thread_data->elements[i])
		{
			localMin = thread_data->elements[i];
		}
		if (localMax < thread_data->elements[i])
		{
			localMax = thread_data->elements[i];
		}
	}

	// uporedi lokalni maksimum sa globalnim maksimumom
	pthread_mutex_lock(thread_data->lock_addr);
	if (localMin < results.minimum)
	{
		results.minimum = localMin;
	}
	if (localMax > results.maximum)
	{
		results.maximum = localMax;
	}
	pthread_mutex_unlock(thread_data->lock_addr);

//  sve što je uslovljeno sa PTW32_STATIC_LIB je potrebno 
//  samo za pthreads-win32 i to samo u static linked varijanti
#ifdef PTW32_STATIC_LIB
	}
	pthread_win32_thread_detach_np();
#endif
	// pthread_exit(NULL); nije neophodno pošto će implicitno biti pozvano
	return 0;
}

int main()
{
//  sve što je uslovljeno sa PTW32_STATIC_LIB je potrebno 
//  samo za pthreads-win32 i to samo u static linked varijanti
#ifdef PTW32_STATIC_LIB
	pthread_win32_process_attach_np();
	pthread_win32_thread_attach_np();
	{
#endif
	// niz i broj elemenata
	int *elements, element_count; 
	unsigned i, slice_size;
	ThreadData thread_data[NUM_THREADS];
	pthread_t workers[NUM_THREADS];
	pthread_attr_t attributes;
	pthread_mutex_t lock;
	
	create_and_load_array(&elements, &element_count);
	if (0 != element_count)
	{
		pthread_attr_init(&attributes);
		pthread_attr_setdetachstate(&attributes, PTHREAD_CREATE_JOINABLE);
		pthread_mutex_init(&lock, NULL);

		results.minimum = results.maximum = elements[0] * elements[0];
		for (i = 0; i < NUM_THREADS; i++)
		{
			thread_data[i].elements = elements;
			thread_data[i].lock_addr = &lock;

			slice_size = element_count / NUM_THREADS;
			thread_data[i].base =  slice_size * i;

			/* poslednja nit dobija eventualne viškove */
			if (NUM_THREADS - 1 == i)
			{
				slice_size += element_count % NUM_THREADS; 
			}
			thread_data[i].end_offset = thread_data[i].base + slice_size;
			pthread_create(&workers[i], NULL, work_routine, (void*) (thread_data + i) );
		}
		// atributi postaju nepotrebni nakon što se stvaranje niti završi
		pthread_attr_destroy(&attributes);

		for (i = 0; i < NUM_THREADS; i++)
		{
			pthread_join(workers[i], NULL);
		}

		print_results(results);

		pthread_mutex_destroy(&lock);
		free(elements);
	}

//  sve što je uslovljeno sa PTW32_STATIC_LIB je potrebno 
//  samo za pthreads-win32 i to samo u static linked varijanti
#ifdef PTW32_STATIC_LIB
	}
	pthread_win32_thread_detach_np();
	pthread_win32_process_detach_np();
#endif
}
